"""Tests of paramattr calibrations applied to properties and methods.

This pattern was used to implement attenuators:
    https://github.com/usnistgov/ssmdevices/blob/main/ssmdevices/instruments/attenuators.py
"""

import pytest
import labbench as lb
from labbench import paramattr as attr
from labbench.testing import store_backend

lb.util.force_full_traceback(True)

@store_backend.key_adapter(defaults={'attenuation_setting': 0})
class PropertyDevice(store_backend.StoreTestDevice):
    frequency: float = attr.value.float(
        5e9,
        allow_none=True,
        min=1e9,
        max=5e9,
        help='frequency for calibration data (None for no calibration)',
        label='Hz',
    )

    output_power_offset: float = attr.value.float(
        default=10.,
        allow_none=True,
        help='output power level at 0 dB attenuation',
        label='dBm',
    )

    calibration_path: float = attr.value.Path(
        default='tests/data/attenuator_cal.csv',
        allow_none=True,
        must_exist=True,
        cache=True,
        help='path to the calibration table csv file (containing frequency ' '(row) and attenuation setting (column))',
    )

    # the only property that requests an attenuation setting in the device
    attenuation_setting = attr.property.float(
        key='attenuation_setting',
        min=0,
        max=110,
        step=5,
        label='dB',
        help='uncalibrated attenuation',
    )

    # the remaining traits are calibration corrections for attenuation_setting
    attenuation = attenuation_setting.corrected_from_table(
        path_attr=calibration_path,
        index_lookup_attr=frequency,
        table_index_column='Frequency(Hz)',
        help='calibrated attenuation',
    )

    output_power = attenuation_setting.corrected_from_expression(
        -attenuation + output_power_offset,
        help='calibrated output power level',
        label='dBm',
    )


@pytest.fixture
def property_device():
    device = PropertyDevice()
    device.open()
    yield device
    device.close()

def test_property_get_through_table(property_device):
    for attenuation_setting in (0,10,100):
        property_device.attenuation_setting = attenuation_setting
        attenuation = property_device.attenuation
        assert attenuation == attenuation_setting + 1, \
               f'attenuation was {attenuation} dB when attenuation setting was set to {attenuation_setting} dB'

def test_property_set_through_table(property_device):
    for attenuation in (1,11,111):
        property_device.attenuation = attenuation
        attenuation_setting = property_device.attenuation_setting
        assert attenuation_setting == attenuation - 1, \
               f'attenuation setting {attenuation_setting} dB when attenuation was set to {attenuation} dB'

    with pytest.raises(ValueError):
        property_device.attenuation = 0

    with pytest.raises(ValueError):
        property_device.attenuation = 112

def test_property_get_with_offset(property_device):
    for attenuation in (1,11,111):
        property_device.attenuation = attenuation
        output_power = property_device.output_power
        assert output_power == property_device.output_power_offset - attenuation, \
               f'output power {output_power} dBm when attenuation was set to {attenuation} dB'

def test_property_set_with_offset(property_device):
    for output_power in (9,-1,-101.):
        property_device.output_power = output_power
        attenuation = property_device.attenuation
        assert output_power == property_device.output_power_offset - attenuation, \
               f'output power {output_power} dBm when attenuation was set to {attenuation} dB'

    with pytest.raises(ValueError):
        property_device.output_power = 10

    with pytest.raises(ValueError):
        property_device.output_power = -102.



class MethodDevice(store_backend.StoreTestDevice):
    frequency: float = attr.value.float(
        5e9,
        allow_none=True,
        min=1e9,
        max=5e9,
        help='frequency for calibration data (None for no calibration)',
        label='Hz',
    )

    output_power_offset: float = attr.value.float(
        default=10.,
        allow_none=True,
        help='output power level at 0 dB attenuation',
        label='dBm',
    )

    calibration_path: float = attr.value.Path(
        default='tests/data/attenuator_cal.csv',
        allow_none=True,
        must_exist=True,
        cache=True,
        help='path to the calibration table csv file (containing frequency ' '(row) and attenuation setting (column))',
    )

    # the only property that requests an attenuation setting in the device
    attenuation_setting = attr.method.float(
        key='attenuation_setting',
        min=0,
        max=110,
        step=5,
        label='dB',
        help='uncalibrated attenuation',
    )

    # the remaining traits are calibration corrections for attenuation_setting
    attenuation = attenuation_setting.corrected_from_table(
        path_attr=calibration_path,
        index_lookup_attr=frequency,
        table_index_column='Frequency(Hz)',
        help='calibrated attenuation',
    )

    output_power = attenuation_setting.corrected_from_expression(
        -attenuation + output_power_offset,
        help='calibrated output power level',
        label='dBm',
    )

@pytest.fixture
def method_device():
    device = MethodDevice()
    device.open()
    yield device
    device.close()


def test_method_get_through_table(method_device):
    for attenuation_setting in (0,10,100):
        method_device.attenuation_setting(attenuation_setting)
        attenuation = method_device.attenuation()
        assert attenuation == attenuation_setting + 1, \
               f'attenuation was {attenuation} dB when attenuation setting was set to {attenuation_setting} dB'

def test_method_set_through_table(method_device):
    for attenuation in (1,11,111):
        method_device.attenuation(attenuation)
        attenuation_setting = method_device.attenuation_setting()
        assert attenuation_setting == attenuation - 1, \
               f'attenuation setting {attenuation_setting} dB when attenuation was set to {attenuation} dB'

    with pytest.raises(ValueError):
        method_device.attenuation(0)

    with pytest.raises(ValueError):
        method_device.attenuation(112)

def test_method_get_with_offset(method_device):
    for attenuation in (1,11,111):
        method_device.attenuation(attenuation)
        output_power = method_device.output_power()
        assert output_power == method_device.output_power_offset - attenuation, \
               f'output power {output_power} dBm when attenuation was set to {attenuation} dB'

def test_method_set_with_offset(method_device):
    for output_power in (9,-1,-101.):
        method_device.output_power(output_power)
        attenuation = method_device.attenuation()
        assert output_power == method_device.output_power_offset - attenuation, \
               f'output power {output_power} dBm when attenuation was set to {attenuation} dB'

    with pytest.raises(ValueError):
        method_device.output_power(10)

    with pytest.raises(ValueError):
        method_device.output_power(-102)



"""
This implementation is deeply intertwined with obscure details of the python object
model. Consider starting with a close read of the documentation and exploring
the objects in an interpreter instead of reverse-engineering this code.
"""

import inspect
import sys
import traceback
from functools import wraps

import typing_extensions as typing

from . import paramattr as attr
from . import util
from .paramattr._bases import (
    HasParamAttrs,
    Undefined,
    hold_attr_notifications,
)


def find_device_instances(depth=1):
    """Look for Device instances, and their names, in the calling
    code context (depth == 1), its caller (depth == 2), and so on.
    Checks locals() in that context first.
    If no Device instances are found there, search the first
    argument of the first function argument, in case this is
    a method in a class.
    """
    from inspect import currentframe

    f = frame = currentframe()
    for i in range(depth):
        f = f.f_back
    try:
        ret = {k: v for k, v in list(f.frame.f_locals.items()) if isinstance(v, Device)}

        # If the context is a function, look in its first argument,
        # in case it is a method. Search its class instance.
        if len(ret) == 0 and len(f.frame.f_code.co_varnames) > 0:
            obj = f.frame.f_locals[f.frame.f_code.co_varnames[0]]
            for k, v in obj.__dict__.items():
                if isinstance(v, Device):
                    ret[k] = v
    finally:
        del f, frame

    return ret


class DisconnectedBackend:
    """ "Null Backend" implementation to raises an exception with discriptive
    messages on attempts to use a backend before a Device is connected.
    """

    def __init__(self, dev):
        """dev may be a class or an object for error feedback"""
        if isinstance(dev, str):
            self.name = dev
        elif getattr(dev, '_owned_name', None) is not None:
            self.name = dev._owned_name
        else:
            self.name = f'{dev.__class__.__qualname__} instance'

    @util.hide_in_traceback
    def __getattr__(self, key):
        msg = f'open {self.name} first to access its backend'
        raise ConnectionError(msg)

    def __repr__(self):
        return 'DisconnectedBackend()'

    def __copy__(self, memo=None):
        return DisconnectedBackend(self.name)

    str = __repr__
    __deepcopy__ = __copy__


def log_paramattr_events(msg):
    """emit debug messages for trait values"""

    if msg['name'] == 'isopen':
        return

    device = msg['owner']
    attr_name = msg['name']

    try:
        attr_def = attr.get_class_attrs(device)[attr_name]
    except KeyError:
        print('attr name: ', msg, attr_name)
        raise

    label = ' '
    if msg['type'] == 'set':
        if attr_def.label:
            label = f' ({attr_def.label})'.rstrip()
        value = repr(msg['new']).rstrip()
        if len(value) > 180:
            value = f'<data of type {type(msg["new"]).__qualname__}>'
        device._logger.debug(f'{value}{label} → {attr_name}')
    elif msg['type'] == 'get':
        if attr_def.label:
            label = f' ({attr_def.label})'
        value = repr(msg['new'])
        if len(value) > 180:
            value = f'<data of type {type(msg["new"]).__qualname__}>'
        device._logger.debug(f'{attr_name} → {value} {label}'.rstrip())
    else:
        device._logger.debug(f'unknown operation type "{msg["type"]}"')


@typing.dataclass_transform(
    kw_only_default=True,
    eq_default=False,
    field_specifiers=attr.value._ALL_TYPES,
)
class DeviceDataClass(HasParamAttrs, util.Ownable):
    @typing.overload
    def __init__(self, **values):
        ...

    def __init__(self, resource=Undefined, **values):
        """Update default values with these arguments on instantiation."""

        if resource is Undefined:
            values['resource'] = type(self).resource.default
        else:
            values['resource'] = resource

        # validate presence of required arguments
        inspect.signature(self.__init__).bind(**values)

        super().__init__()

        with hold_attr_notifications(self):
            for name, init_value in values.items():
                setattr(self, name, init_value)

            other_names = set(self._attr_defs.value_names()) - set(values.keys())
            for name in other_names:
                attr_def = getattr(type(self), name)
                if isinstance(attr_def, attr.value.Value):
                    if attr_def.default is not Undefined:
                        self._attr_store.cache[name] = attr_def.default
                    elif attr_def.allow_none:
                        self._attr_store.cache[name] = None
                    else:
                        attr_desc = attr.__repr__(owner_inst=self)
                        raise TypeError(
                            f'unable to determine an initial value for {attr_desc} - define it with allow_none=True or default=<default value>'
                        )

        util.Ownable.__init__(self)

        self.backend = DisconnectedBackend(self)

        # Instantiate property trait now. It needed to wait until this point, after values are fully
        # instantiated, in case property trait implementation depends on values
        self.open = self.__open_wrapper__
        self.close = self.__close_wrapper__

    @classmethod
    @util.hide_in_traceback
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.__annotations__ = typing.get_type_hints(cls)
        cls.__update_signature__()

    @classmethod
    @util.hide_in_traceback
    def __update_signature__(cls):
        # Generate a signature for documentation and code autocomplete
        params = [
            inspect.Parameter('self', kind=inspect.Parameter.POSITIONAL_OR_KEYWORD),
            inspect.Parameter(
                'resource',
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=cls.resource.default,
                annotation=cls.resource._type,
            ),
        ]

        # generate and apply the sequence of call signature parameters
        constructor_attrs = []
        for name in cls.__annotations__.keys():
            attr_def = getattr(cls, name)
            constructor_attrs.append(attr_def)

            if not isinstance(attr_def, attr.value.Value):
                annot_desc = f'{name}: {cls.__annotations__[name].__name__}'
                wrong_type = type(attr_def)
                raise TypeError(
                    f'only labbench.paramattr.value descriptors may be annotated in labbench Device classes, but "{annot_desc}" annotates {wrong_type!r}'
                )

            elif name == 'resource':
                # defined above for its POSITIONAL_OR_KEYWORD special casing
                continue

            else:
                if attr_def.only and sys.version_info > (3, 10):
                    # Union[*attr_def.only] is sooo close
                    annotation = typing.Union.__getitem__(tuple(attr_def.only))
                else:
                    annotation = attr_def._type

                if attr_def.allow_none:
                    annotation = typing.Union[annotation, None]

                params.append(
                    inspect.Parameter(
                        name,
                        kind=inspect.Parameter.KEYWORD_ONLY,
                        default=attr_def.default,
                        annotation=annotation,
                    )
                )

        # we need a wrapper so that __init__ can be modified separately for each subclass
        cls.__init__ = util.copy_func(cls.__init__)
        cls.__init__.__signature__ = inspect.Signature(params)

        # generate the __init__ docstring
        value_docs = ''.join([f'    {t.name}: {t.doc(as_argument=True)}\n' for t in constructor_attrs])
        cls.__init__.__doc__ = f'\nArguments:\n{value_docs}'


class Device(DeviceDataClass):
    r"""base class for labbench device wrappers.

    Drivers that subclass `Device` share

    * standardized connection management via context blocks (the `with` statement)
    * hooks for automatic data logging and heads-up displays
    * API style consistency
    * bounds checking and casting for typed attributes

    .. note::
        This `Device` base class has convenience
        functions for device control, but no implementation.

        Some wrappers for particular APIs labbench Device subclasses:

            * VISADevice: pyvisa,
            * ShellBackend: binaries and scripts
            * Serial: pyserial
            * DotNetDevice: pythonnet

        (and others). If you are implementing a driver that uses one of
        these backends, inherit from the corresponding class above, not
        `Device`.

    """

    resource: str = attr.value.str(
        default=None, allow_none=True, cache=True, kw_only=False, help='device address or URI'
    )

    """ Container for property trait traits in a Device. Getting or setting property trait traits
        triggers live updates: communication with the device to get or set the
        value on the Device. Therefore, getting or setting property trait traits
        needs the device to be connected.

        To set a property trait value inside the device, use normal python assigment::

            device.parameter = value

        To get a property trait value from the device, you can also use it as a normal python variable::

            variable = device.parameter + 1
    """
    backend = DisconnectedBackend(None)
    """ .. this attribute is some reference to a controller for the device.
        it is to be set in `connect` and `disconnect` by the subclass that implements the backend.
    """

    # Backend classes may optionally overload these, and do not need to call the parents
    # defined here
    def open(self):
        """Backend implementations overload this to open a backend
        connection to the resource.
        """
        attr.observe(self, log_paramattr_events)

    def close(self):
        """Backend implementations must overload this to disconnect an
        existing connection to the resource encapsulated in the object.
        """
        self.backend = DisconnectedBackend(self)
        self.isopen
        attr.unobserve(self, log_paramattr_events)

    @util.hide_in_traceback
    @wraps(open)
    def __open_wrapper__(self):
        """A wrapper for the connect() method. It steps through the
        method resolution order of self.__class__ and invokes each open()
        method, starting with labbench.Device and working down
        """
        if self.isopen:
            self._logger.debug(f'attempt to open {self}, which is already open')
            return

        self.backend = None

        try:
            for opener in util.find_methods_in_mro(self.__class__, 'open', Device)[::-1]:
                opener(self)
        except BaseException:
            self.backend = DisconnectedBackend(self)
            raise

        self._logger.debug('opened')

        # Force an update to self.isopen
        self.isopen

    @util.hide_in_traceback
    @wraps(close)
    def __close_wrapper__(self):
        """A wrapper for the close() method that runs
        cleanup() before calling close(). close()
        will still be called if there is an exception in cleanup().
        """
        # Try to run cleanup(), but make sure to run
        # close() even if it fails
        # with self._hold_notifications('isopen'):
        if not self.isopen:
            return

        methods = util.find_methods_in_mro(self.__class__, 'close', Device)

        all_ex = []
        for closer in methods:
            try:
                closer(self)
            except BaseException:
                all_ex.append(sys.exc_info())

        try:
            # Print tracebacks for any suppressed exceptions
            for ex in all_ex[::-1]:
                # If ThreadEndedByMaster was raised, assume the error handling in
                # util.concurrently will print the error message
                if ex[0] is not util.ThreadEndedByMaster:
                    depth = len(tuple(traceback.walk_tb(ex[2])))
                    traceback.print_exception(*ex, limit=-(depth - 1))
                    sys.stderr.write('(Exception suppressed to continue close)\n\n')

            self.isopen

            self._logger.debug('closed')
        finally:
            if len(all_ex) > 0:
                ex = util.ConcurrentException(f'multiple exceptions while closing {self}')
                ex.thread_exceptions = all_ex
                raise ex

    @util.hide_in_traceback
    def __enter__(self):
        try:
            self.open()
            return self
        except BaseException as e:
            args = list(e.args)
            if len(args) > 0:
                args[0] = f'{self!r}: {args[0]}'
                e.args = tuple(args)
            raise e

    @util.hide_in_traceback
    def __exit__(self, type_, value, traceback):
        try:
            self.close()
        except BaseException as e:
            args = list(e.args)
            args[0] = f'{self!r}: {args[0]!s}'
            e.args = tuple(args)
            raise e

    # Object boilerplate
    def __del__(self):
        try:
            isopen = self.isopen
        except (AttributeError, KeyError):
            # the object failed to instantiate properly
            isopen = False

        if isopen:
            self.close()

    def __repr__(self):
        name = self.__class__.__qualname__
        if hasattr(self, 'resource'):
            if self.resource != type(self).resource.default:
                resource_str = repr(self.resource)
            else:
                resource_str = ''
            return f'{name}({resource_str})'
        else:
            # In case an exception has occurred before __init__
            return f'{name}()'

    @attr.property.bool()
    def isopen(self):
        """`True` if the backend is ready for use"""
        try:
            return DisconnectedBackend not in self.backend.__class__.__mro__
        except BaseException:
            # Run into this sometimes on reloading a module or ipython shell:
            # the namespace is gone. we just assume disconnected
            return False


Device.__init_subclass__()



import datetime
import email.mime.text
import io
import json
import logging
import socket
import sys
import time
from traceback import format_exc, format_exception_only, format_tb

from . import _device as core
from . import paramattr as attr
from . import util

try:
    git = util.lazy_import('git')
    pd = util.lazy_import('pandas')
    pip = util.lazy_import('pip')
    smtplib = util.lazy_import('smtplib')
except RuntimeWarning:
    # not executed: help coding tools recognize lazy_imports as imports
    import smtplib

    import git
    import pandas as pd
    import pip

__all__ = ['Host', 'Email']


class LogStreamBuffer:
    def __init__(self):
        self._value = ''

    def write(self, msg):
        self._value = self._value + msg
        return len(msg)

    def read(self):
        ret, self._value = self._value, ''
        return ret

    def flush(self):
        pass


class LogStderr(core.Device):
    """This "Device" logs a copy of messages on sys.stderr while connected."""

    log = ''

    def open(self):
        self._stderr, self._buf, sys.stderr = sys.stderr, io.StringIO(), self

    def close(self):
        if self.isopen:
            sys.stderr = self._stderr
            self.log = self._buf.getvalue()
            self._buf.close()

    def write(self, what):
        self._stderr.write(what)
        self._buf.write(what)

    def flush(self):
        try:
            self._buf.flush()
            self.log += self._buf.getvalue()
        except ValueError:
            pass


class Email(core.Device):
    """Sends a notification message on disconnection. If an exception
    was thrown, this is a failure subject line with traceback information
    in the main body. Otherwise, the message is a success message in the
    subject line. Stderr is also sent.
    """

    resource: str = attr.value.NetworkAddress(default='smtp.nist.gov', help='smtp server to use', cache=True)
    port: int = attr.value.int(default=25, min=1, help='TCP/IP port', cache=True)
    sender: str = attr.value.str(default='myemail@nist.gov', help='email address of the sender', cache=True)
    recipients: list = attr.value.list(
        default=['myemail@nist.gov'], help='list of email addresses of recipients', cache=True
    )

    success_message: str = attr.value.str(
        default='Test finished normally',
        allow_none=True,
        help='subject line for test success emails (None to suppress the emails)',
        cache=True,
    )

    failure_message: str = attr.value.str(
        default='Exception ended test early',
        allow_none=True,
        help='subject line for test failure emails (None to suppress the emails)',
        cache=True,
    )

    def _send(self, subject, body):
        sys.stderr.flush()
        self.backend.flush()

        msg = email.mime.text.MIMEText(body, 'html')
        msg['From'] = self.sender
        msg['Subject'] = subject
        msg['To'] = ', '.join(self.recipients)
        self.server = smtplib.SMTP(self.resource, self.port)

        try:
            self.server.sendmail(self.sender, self.recipients, msg.as_string())
        finally:
            self.server.quit()

    def open(self):
        self.backend = LogStderr()
        self.backend.open()

    def close(self):
        if self.isopen:
            self.backend.close()
            util.sleep(1)
            self.send_summary()

    def send_summary(self):
        """Send the email containing the final property trait of the test."""
        exc = sys.exc_info()

        if exc[0] is KeyboardInterrupt:
            return

        if exc != (None, None, None):
            if self.failure_message is None:
                return
            subject = self.failure_message
            message = '<b>Exception</b>\n' + '<font face="Courier New, Courier, monospace">' + format_exc() + '</font>'
        else:
            if self.success_message is None:
                return
            subject = self.success_message
            message = ''

        if len(self.backend.log) > 0:
            message = (
                message
                + '\n\n<b>Standard error</b>\n'
                + '<font face="Courier New, Courier, monospace">'
                + self.backend.log
                + '</font>'
            )

        self._send(subject, message.replace('\n', '<br>'))

        return subject, message


class JSONFormatter(logging.Formatter):
    _last = []

    def __init__(self, *args, **kws):
        super().__init__(*args, **kws)
        self.t0 = time.time()

    @staticmethod
    def json_serialize_dates(obj):
        """JSON serializer for objects not serializable by default json code"""

        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        raise TypeError(f'Type {type(obj).__qualname__} not serializable')

    def format(self, rec):
        """Return a YAML string for each logger record"""

        # if isinstance(rec, core.Device):
        #     log_prefix = rec._owned_name.replace(".", ",")
        # else:
        #     log_prefix = ''

        # object = getattr(rec, 'object', None)

        msg = dict(
            message=rec.msg,
            time=datetime.datetime.fromtimestamp(rec.created),
            elapsed_seconds=rec.created - self.t0,
            level=rec.levelname,
            object=getattr(rec, 'object', None),
            object_log_name=getattr(rec, 'owned_name', None),
            source_file=rec.pathname,
            source_line=rec.lineno,
            process=rec.process,
            thread=rec.threadName,
        )

        if rec.threadName != 'MainThread':
            msg['thread'] = rec.threadName

        etype, einst, exc_tb = sys.exc_info()
        if etype is not None:
            msg['exception'] = format_exception_only(etype, einst)[0].rstrip()
            msg['traceback'] = ''.join(format_tb(exc_tb)).splitlines()

        self._last.append((rec, msg))

        return json.dumps(msg, indent=True, default=self.json_serialize_dates) + ','


class Host(core.Device):
    # Settings
    git_commit_in: str = attr.value.str(
        default=None,
        allow_none=True,
        help='git commit on open() if run inside a git repo with this branch name',
    )

    time_format = '%Y-%m-%d %H:%M:%S'

    def open(self):
        """The host setup method tries to commit current changes to the tree"""

        # touch git to ensure completed import
        git.__version__

        log_formatter = JSONFormatter()
        stream = LogStreamBuffer()
        sh = logging.StreamHandler(stream)
        sh.setFormatter(log_formatter)
        sh.setLevel(logging.DEBUG)

        # Add to the labbench logger handler
        logger = logging.getLogger('labbench')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(sh)

        # git repository information
        try:
            repo = git.Repo('.', search_parent_directories=True)
            self._logger.debug('running in git repository')
            if self.git_commit_in is not None and repo.active_branch == self.git_commit_in:
                repo.index.commit('start of measurement')
                self._logger.debug('git commit finished')
        except git.NoSuchPathError:
            repo = None
            self._logger.info('not running in a git repository')

        self.backend = {
            'logger': logger,
            'log_stream': stream,
            'log_handler': sh,
            'log_formatter': log_formatter,
            'repo': repo,
        }

        # Preload the git repo parameters
        for name in attr.get_class_attrs(self).keys():
            if name.startswith('git'):
                getattr(self, name)

    def close(self):
        try:
            self.backend['logger'].removeHandler(self.backend['log_handler'])
        except (AttributeError, TypeError):
            pass
        try:
            self.backend['log_stream'].close()
        except (AttributeError, TypeError):
            pass

    def metadata(self):
        """Generate the metadata associated with the host and python distribution"""
        return dict(python_modules=self.__python_module_versions())

    def __python_module_versions(self):
        """Enumerate the versions of installed python modules"""

        versions = dict([str(d).lower().split(' ') for d in pip.get_installed_distributions()])
        running = dict(sorted([(k, versions[k.lower()]) for k in sys.modules.keys() if k in versions]))
        return pd.Series(running).sort_index()

    @attr.property.str()
    def time(self):
        """Get a timestamp of the current time"""
        now = datetime.datetime.now()
        return f'{now.strftime(self.time_format)}.{now.microsecond}'

    @attr.property.list()
    def log(self):
        """Get the current host log contents."""
        self.backend['log_handler'].flush()
        txt = self.backend['log_stream'].read()
        if len(txt) > 1:
            self._txt = txt
            self._serialized = '[' + txt[:-2] + ']'
            self._ret = json.loads(self._serialized)  # self.backend['log_stream'].read().replace('\n', '\r\n')
            return self._ret
        else:
            return {}

    @attr.property.str(cache=True)
    def git_commit_id(self):
        """Try to determine the current commit hash of the current git repo"""

        try:
            commit = self.backend['repo'].commit()
            return commit.hexsha
        except git.NoSuchPathError:
            return ''

    @attr.property.str(cache=True)
    def git_remote_url(self):
        """Try to identify the remote URL of the repository of the current git repo"""
        try:
            return next(self.backend['repo'].remote().urls)
        except BaseException:
            return ''

    @attr.property.str(cache=True)
    def hostname(self):
        """Get the name of the current host"""
        return socket.gethostname()

    @attr.property.str(cache=True)
    def git_browse_url(self):
        """URL for browsing the current git repository"""
        return f'{self.git_remote_url}/tree/{self.git_commit_id}'

    @attr.property.str(cache=True)
    def git_pending_changes(self):
        if self.backend['repo'] is not None:
            diffs = self.backend['repo'].index.diff(None)
            return str(tuple(diff.b_path for diff in diffs))[1:-1]
        else:
            return ''


if __name__ == '__main__':
    #    core.show_messages('DEBUG')
    #
    #    with Host() as pc:
    #        print(pc.time)

    with Email(recipients=['daniel.kuester@nist.gov']) as mail:
        pass



import pytest
from paramattr_tooling import eval_set_then_get, loop_closed_loop_set_gets

import labbench as lb
from labbench import paramattr as attr
from labbench.testing import store_backend


@store_backend.key_adapter(defaults={'str_or_none': None, 'str_cached': 'cached string'})
class StoreTestDevice(store_backend.StoreTestDevice):
    LOOP_TEST_VALUES = {
        # make sure all test values conform to these general test values
        int: 5,
        float: 3.14,
        str: 'hi',
        bool: True,
        object: None,
    }

    # test both getting and setting
    bool_keyed = attr.property.bool(key='bool_keyed')
    int_keyed_unbounded = attr.property.int(key='int_keyed_unbounded')

    int_decorated_low_bound_getonly = attr.property.int(min=0, sets=False)

    @int_decorated_low_bound_getonly.getter
    def _(self):
        return self.backend.setdefault('int_decorated_low_bound_getonly', 0)

    @attr.property.int(min=10, gets=False).setter
    def int_decorated_low_bound_setonly(self, set_value=lb.Undefined):
        self.backend['int_decorated_high_bound_setonly'] = set_value

    str_or_none = attr.property.str(key='str_or_none', allow_none=True)
    str_cached = attr.property.str(key='str_cached', cache=True)
    any = attr.property.any(key='any', allow_none=True)


def set_param(device, attr_name, value, arguments={}):
    if len(arguments) > 0:
        raise ValueError('properties do not accept arguments')
    setattr(device, attr_name, value)


def get_param(device, attr_name, arguments={}):
    if len(arguments) > 0:
        raise ValueError('properties do not accept arguments')
    return getattr(device, attr_name)


def set_then_get(device, attr_name, value_in, arguments={}):
    set_param(device, attr_name, value_in, arguments)
    return get_param(device, attr_name, arguments)


#
# Fixtures that convert to arguments in test functions
#
@pytest.fixture(autouse=True, scope='module')
def labbench_fixture():
    lb.visa_default_resource_manager('@sim')
    lb.show_messages('info')
    lb.util.force_full_traceback(True)


@pytest.fixture
def role_type():
    return lb.paramattr.property.Property


@pytest.fixture
def opened_device():
    device = StoreTestDevice()
    device.open()
    yield device
    device.close()


@pytest.fixture
def instantiated_device():
    device = StoreTestDevice()
    device.open()
    yield device
    device.close()


#
# The tests
#
def test_basic_get(opened_device):
    _ = opened_device.any


def test_basic_set(opened_device):
    opened_device.any = 5
    assert opened_device.any == 5


def test_cache(opened_device):
    # repeat to set->get to ensure proper caching
    eval_set_then_get(opened_device, 'str_cached', set_then_get)
    result = eval_set_then_get(opened_device, 'str_cached', set_then_get)

    assert result['get_count'] == 0, 'cache test - second "get" operation count'

    assert result['set_count'] == 2, 'cache test - second "get" operation count'


def test_all_get_sets(opened_device, role_type):
    loop_closed_loop_set_gets(opened_device, role_type, set_then_get)



"""functions and CLI tools for mapping labbench objects onto config directories"""

import inspect
import os
from numbers import Number
from pathlib import Path

from . import util
from ._rack import Rack, import_as_rack, update_parameter_dict

# some packages install ruamel_yaml, others ruamel.yaml. fall back to ruamel_yaml in case ruamel.yaml fails
# using ruamel yaml instead of pyyaml because it allows us to place comments for human readability
try:
    import ruamel.yaml as ruamel_yaml
    from ruamel.yaml.comments import CommentedMap
except ModuleNotFoundError:
    import ruamel_yaml
    from ruamel_yaml.comments import CommentedMap

_yaml = ruamel_yaml.YAML()
_yaml.indent(mapping=4, sequence=4)

RACK_CONFIG_FILENAME = 'config.yaml'
EMPTY = inspect.Parameter.empty

# csv files that define sequences for function execution
INDEX_COLUMN_NAME = 'step_name'

_FIELD_SOURCE = 'source'
_FIELD_DEVICES = 'devices'
_FIELD_KEYWORD_DEFAULTS = 'default_arguments'


def _yaml_comment_out(cm, key):
    """comment out the line containing the item with the specified key"""

    from ruamel_yaml.error import CommentMark
    from ruamel_yaml.tokens import CommentToken

    cm.ca.items.setdefault(key, [None, [], None, None])
    cm.ca.items[key][1] = [CommentToken('# ', CommentMark(0), None)]


def _quote_strings_recursive(cm):
    """apply quotes to dict values that have str type"""

    from ruamel_yaml.scalarstring import DoubleQuotedScalarString as quote

    ret = dict()

    for k, v in cm.items():
        if isinstance(v, str):
            ret[k] = quote(v)
        elif isinstance(v, dict):
            ret[k] = _quote_strings_recursive(v)

    return ret


def _search_method_parameters(rack_cls_or_obj):
    """finds parameters of methods in rack_cls and its children recursively.

    Arguments:
        A subclass of Rack

    Returns:
        {<name (str)>: <(inspect.Parameter)>}, {<name (str)>: {<short_name (str)>: <callable method>>}}
    """

    parameters = {}
    methods = {}

    for rack in rack_cls_or_obj._owners.values():
        # recurse into child racks
        if not isinstance(rack, Rack):
            continue

        p, m = _search_method_parameters(rack)

        update_parameter_dict(parameters, p)

        for name, callables in m.items():
            methods.setdefault(name, {}).update(callables)

    if inspect.isclass(rack_cls_or_obj):
        rack = rack_cls_or_obj()
    else:
        rack = rack_cls_or_obj

    for method in rack._methods.values():
        p = dict(list(method.extended_signature().parameters.items()))  # skip 'self'

        update_parameter_dict(parameters, p)

        short_names = list(method.__call__.__signature__.parameters.keys())[1:]  # skip 'self'
        long_names = list(p.keys())  # extended_signature() does not include 'self'
        for short_name, long_name in zip(short_names, long_names):
            methods.setdefault(long_name, {})[short_name] = method

    return parameters, methods


def _adjust_sequence_defaults(rack_cls: type, defaults_in: dict, **override_defaults):
    """adjusts the method argument parameters in the Rack subclass `cls` according to config file"""
    params, methods = _search_method_parameters(rack_cls)

    defaults_in = dict(defaults_in, **override_defaults)

    # sequences = [
    #     obj for obj in rack_cls._ownables.values() if isinstance(obj, Sequence)
    # ]

    for name, default in dict(defaults_in).items():
        if default == params[name].default:
            del defaults_in[name]
            continue

        annot = params[name].annotation
        if name not in methods:
            clsname = rack_cls.__qualname__
            raise KeyError(f"'{name}' is not a keyword argument of any method of '{clsname}'")

        elif annot is not EMPTY and not isinstance(default, annot):
            if isinstance(default, Number) and issubclass(annot, Number):
                # allow casting for numbers
                default = annot(default)

            else:
                raise TypeError(
                    f"the keyword default configuration at key '{name}' with value "
                    f"'{default}' conflicts with annotated type '{annot.__qualname__}'"
                )

        # update the call signature
        for short_name, method in methods[name].items():
            method.set_kwdefault(short_name, default)

    if len(defaults_in) > 0:
        util.logger.debug(f'applied defaults {defaults_in}')


def write_table_stub(rack: Rack, name: str, path: Path, with_defaults: bool = False):
    """forms an empty DataFrame containing the headers needed for Sequence
    csv files.

    Arguments:
        rack: the Rack instance containing the sequence
        path: base directory where the csv should be saved
        with_defaults: whether to include columns when method parameters have defaults

    """

    import pandas as pd

    func = getattr(rack, name)
    if not callable(func):
        raise TypeError(f'{func} is not callable')
    try:
        sig = inspect.signature(func)
    except ValueError:
        sig = inspect.signature(func.__call__)

    # pick out the desired column names based on with_defaults
    params = sig.parameters
    columns = [name for name, param in list(params.items())[1:] if with_defaults or param.default is EMPTY]

    if with_defaults:
        defaults = [[None if params[name].default is EMPTY else params[name].default for name in columns]]
    else:
        defaults = []

    df = pd.DataFrame(defaults, columns=columns)
    df.index.name = INDEX_COLUMN_NAME
    df.to_csv(path)
    util.logger.debug(f'writing csv template to {path!r}')


def _map_method_defaults(rack_cls):
    params, _ = _search_method_parameters(rack_cls)
    cm = CommentedMap({k: (None if param.default is EMPTY else param.default) for k, param in params.items()})

    for i, k in enumerate(list(cm.keys())[::-1]):
        if params[k].default is EMPTY:
            # comment out lines with no default to distinguish
            # from None value (which is an empty line in yaml)
            _yaml_comment_out(cm, k)

        if params[k].annotation is not EMPTY:
            # comment the type
            cm.yaml_add_eol_comment(key=k, comment=str(params[k].annotation.__name__))

    return cm


def _map_devices(cls):
    cm = CommentedMap()

    for dev_name, dev in cls._devices.items():
        cm[dev_name] = CommentedMap()
        cm.yaml_set_comment_before_after_key(
            dev_name,
            before='\n',
        )

        for value_name in param.list_value_attrs(dev):
            if not param.get_class_attrs(dev)[value_name].sets:
                # only show settable traits
                continue

            cm[dev_name][value_name] = getattr(dev, value_name)
            trait = getattr(type(dev), value_name)

            if trait.help:
                comment = '\n' + trait.help
            else:
                comment = '\n(define this value with help to autogenerate this comment)'

            cm[dev_name].yaml_set_comment_before_after_key(value_name, before=comment, indent=8)

            if trait.type is not None:
                comment = trait.type.__name__
                if trait.label:
                    comment = f'{comment} ({trait.label})'

                cm[dev_name].yaml_add_eol_comment(key=value_name, comment=comment)

    return cm


def dump_rack(
    rack: Rack,
    output_path: Path,
    sourcepath: Path,
    pythonpath: Path = None,
    exist_ok: bool = False,
    with_defaults: bool = False,
    skip_tables: bool = False,
):
    if not isinstance(rack, Rack):
        raise TypeError("'rack' argument must be an instance of labbench.Rack")

    cls = type(rack)

    output_path = Path(output_path)
    output_path.mkdir(exist_ok=exist_ok, parents=True)

    with open(output_path / RACK_CONFIG_FILENAME, 'w') as stream:
        cm = CommentedMap(
            {
                _FIELD_SOURCE: dict(
                    import_string=str(sourcepath),
                    class_name=None if cls.__name__ == '_as_rack' else cls.__name__,
                    python_path=str(pythonpath),
                ),
                _FIELD_KEYWORD_DEFAULTS: _map_method_defaults(rack),
                _FIELD_DEVICES: _map_devices(cls),
            }
        )

        cm.yaml_set_comment_before_after_key(
            _FIELD_SOURCE,
            before='orient the python interpreter to the source',
        )

        cm.yaml_set_comment_before_after_key(
            _FIELD_KEYWORD_DEFAULTS,
            before='\nparameter defaults for sequences in rack:'
            '\nthese parameters can be omitted from sequence table columns',
        )

        cm.yaml_set_comment_before_after_key(
            _FIELD_DEVICES, before='\ndevice settings: initial values for value traits'
        )

        # cm = _quote_strings_recursive(cm)

        _yaml.dump(cm, stream)

    if not skip_tables:
        for name, obj in rack.__dict__.items():
            if not callable(obj):
                continue

            table_path = getattr(obj, '_tags', {}).get('table_path', None)

            if table_path is None and not hasattr(Rack, name):
                table_path = name + '.csv'

            if table_path is not None:
                # write_csv_template(obj, output_path/table_path)
                # obj.to_template(output_path / f"{obj.__name__}.csv")
                write_table_stub(rack, name, output_path / table_path, with_defaults=with_defaults)


def read_yaml_config(config_path: str):
    with open(config_path) as f:
        config = _yaml.load(f)
        util.logger.debug(f'loaded configuration from "{config_path!s}"')
    return config


def load_rack(output_path: str, defaults: dict = {}, apply: bool = True) -> Rack:
    """instantiates a Rack object from a config directory created by dump_rack.

    After instantiation, the current working directory is changed to output_path.
    """

    config_path = Path(output_path) / RACK_CONFIG_FILENAME
    config = read_yaml_config(config_path)

    if 'import_string' not in config[_FIELD_SOURCE]:
        raise KeyError(f"import_string missing from '{config_path!s}'")

    append_path = config[_FIELD_SOURCE]['python_path']

    # synthesize a Rack class
    rack_cls = import_as_rack(
        import_string=config[_FIELD_SOURCE]['import_string'],
        cls_name=config[_FIELD_SOURCE]['class_name'],
        append_path=[] if append_path is None else [append_path],
        # TODO: support extensions to python path?
    )

    if apply:
        os.chdir(output_path)
        _adjust_sequence_defaults(rack_cls, config[_FIELD_KEYWORD_DEFAULTS], **defaults)

    rack_cls._propagate_ownership()

    obj = rack_cls()

    if apply:
        for name, params in config[_FIELD_DEVICES].items():
            try:
                owned_obj = getattr(obj, name)
            except AttributeError:
                objname = type(obj).__qualname__
                raise OSError(f"{config_path} refers to a device '{name}' that does not exist in {objname}")

            for param_name, param_value in params.items():
                setattr(owned_obj, param_name, param_value)

    return obj



import labbench as lb
from labbench import paramattr as attr

flag_start = False


class Shell_Python(lb.ShellBackend):
    binary_path = attr.value.str('python')

    path = attr.value.str(key=None, help='path to a python script file')

    command = attr.value.str(key='-c', help='execute a python command')

    def _flag_names(self):
        return (name for name, trait in self._traits.items() if trait.key is not lb.Undefined)

    def _commandline(self, **flags):
        """Form a list of commandline argument strings for foreground
        or background calls.

        :returns: tuple of string
        """

        # validate the set of flags
        unsupported = set(flags).difference(self._flag_names())
        if len(unsupported) > 1:
            raise KeyError(f'flags {unsupported} are unsupported')

        # apply flags
        for name, value in flags.items():
            setattr(self, name, value)

        cmd = (self.binary_path,)

        # Update property trait traits with the flags
        for name in self._flag_names():
            trait = self[name]
            value = getattr(self, name)
            print(name, repr(value), repr(trait.key), repr(lb.Undefined))

            if value is None:
                continue
            elif trait.key in (None, ''):
                cmd = cmd + (value,)
            elif not isinstance(trait.key, str) and trait.key is not lb._traits.Undefined:
                raise TypeError(f'keys defined in {self} must be strings')
            else:
                cmd = cmd + (trait.key, value)
        return cmd


# class TestSettings(unittest.TestCase):
#     def test_defaults(self):
#         with Mock() as m:
#             for name, trait in m._traits.items():
#                 self.assertEqual(getattr(m, name),
#                                  trait.default, msg=f'defaults: {name}')
#
if __name__ == '__main__':
    import inspect

    lb.show_messages('debug')

    python = Shell_Python(path=r'c:\script.py', command='print("hello world")')
    print(python['resource'].key is lb.Undefined)
    print(lb.Unicode.__defaults__)
    print(inspect.signature(lb.Unicode.__init__))

    print(python._commandline())



import labbench as lb
from labbench import paramattr as attr


def has_steps(attr: attr.ParamAttr):
    return getattr(attr, 'step', None) is not None


def eval_access(device: lb.Device, attr_name, arguments: dict = {}) -> dict:
    attr_def = getattr(type(device), attr_name)
    backend_key = device.backend.get_backend_key(device, attr_def, arguments)
    notifications = [n for n in device.backend.notifications if n['name'] == attr_name]

    return {
        'get_count': device.backend.get_count[backend_key],
        'set_count': device.backend.set_count[backend_key],
        'notifications': notifications,
    }


def eval_set_then_get(device, attr_name, single_set_get: callable, value_in=lb.Undefined, arguments={}):
    attr_def = getattr(type(device), attr_name)

    if value_in is lb.Undefined:
        value_in = device.LOOP_TEST_VALUES[attr_def._type]

    value_out = single_set_get(device, attr_name, value_in, arguments)
    access = eval_access(device, attr_name, arguments)

    return dict(value_in=value_in, value_out=value_out, **access)


def loop_closed_loop_set_gets(device: lb.Device, role_type: type[lb.paramattr.ParamAttr], single_set_get: callable):
    def want_to_set_get(attr_def):
        return (
            isinstance(attr_def, role_type)
            and attr_def.sets
            and attr_def.gets
            and not hasattr(lb.Device, attr_def.name)
            and not has_steps(attr_def)  # steps can make set != get
        )

    attrs = {name: attr_def for name, attr_def in device.get_attr_defs().items() if want_to_set_get(attr_def)}

    for attr_name, attr_def in attrs.items():
        if isinstance(attr_def, attr.method.Method):
            # skip methods with arguments for now
            if len(attr_def.get_kwarg_names()) > 0:
                continue
        test_name = f'{attr_def.ROLE} "{attr_name}"'
        has_reduced_access_count = attr_def.cache or isinstance(attr_def, attr.value.Value)

        device.backend.clear_counts()

        result = eval_set_then_get(device, attr_name, single_set_get)

        assert result['value_in'] == result['value_out'], f'{test_name} - set-get input and output values'

        assert (
            len(result['notifications']) == 1 if has_reduced_access_count else 2
        ), f'{test_name} - callback notification count'

        if isinstance(attr_def, attr.value.Value):
            if len(result['notifications']) > 1:
                assert (
                    result['notifications'][0]['old'] == attr_def.default
                ), f"{test_name} - callback notification prior value for 'set'"
        else:
            assert (
                result['notifications'][0]['old'] == lb.Undefined
            ), f"{test_name} - callback notification prior value for 'set'"

        if not attr_def.cache and len(result['notifications']) > 1:
            assert (
                result['notifications'][1]['old'] == result['value_in']
            ), f"{test_name} - callback notification prior value for 'get'"

        # make sure there weren't any unecessary extra 'get' operations
        assert result['get_count'] == 0 if has_reduced_access_count else 1, f'{test_name} - "get" notification count'
        assert (
            result['set_count'] == 0 if isinstance(attr_def, attr.value.Value) else 1
        ), f'{test_name} - "set" notification count'



import importlib
import sys
import typing
from pathlib import Path

import astor

from labbench import Device, Rack, util
from labbench.paramattr._bases import (
    Any,
    Method,
    ParamAttr,
    T,
    Undefined,
    get_class_attrs,
    list_value_attrs,
)

util.force_full_traceback(True)

VALID_PARENTS = Device, Rack, ParamAttr

import ast
import typing
from inspect import isclass

from ast_decompiler import decompile

import labbench as lb
from labbench import Device, Rack


def nameit(obj):
    if obj is Undefined:
        return 'Undefined'
    if isinstance(obj, typing.GenericAlias):
        return repr(obj)
    elif hasattr(obj, '__name__'):
        return obj.__name__
    if hasattr(obj, '__qualname__'):
        return obj.__qualname__

    t = type(obj)
    if hasattr(t, '__name__'):
        return t.__name__
    elif hasattr(t, '__qualname__'):
        return t.__qualname__

    raise TypeError(f"couldn't name {obj}")


def parse_literal(s: str):
    obj = ast.parse(s).body[0].value
    # obj.ctx = context_source
    return obj


def ast_name(name):
    return ast.Name(id=name, kind=None)


def ast_arg(name, annotation=lb.Undefined):
    annotation = ast_name(annotation) if annotation is not None else None
    return ast.arg(arg=name, annotation=annotation, type_comment=None)


def ast_typehint_optional(t, remap={}):
    remap[type(None)] = None
    args = ','.join([getattr(remap.get(sub, sub), '__name__', sub.__class__.__name__) for sub in typing.get_args(t)])
    print('***', f'_typing.Optional[{args}]')
    return parse_literal(f'_typing.Optional[{args}]')


def ast_signature(args, defaults, annotations):
    annotations = {k: nameit(v) for k, v in annotations.items() if v is not Undefined}

    return ast.arguments(
        posonlyargs=[],
        kw_defaults=[],
        args=[ast_arg(a, annotations.get(a, None)) for a in args],
        vararg=None,
        kwarg=None,  # ast.arg(arg='values', annotation=ast.Name(id='Any', ctx=ast.Load())),
        defaults=[ast_name(repr(d)) for d in defaults if d is not lb.Undefined],
    )


def ast_function_stub(name, args, defaults, annotations, decorator_list=[]):
    if 'return' in annotations:
        if isinstance(annotations['return'], (ast.Attribute, ast.Subscript)):
            type_comment = returns = annotations['return']
        else:
            type_comment = returns = ast_name(nameit(annotations['return']))
    else:
        type_comment = returns = None

    return ast.FunctionDef(
        name=name,
        args=ast_signature(args, defaults, annotations),
        body=[ast.Expr(value=ast.Constant(value=..., kind=None))],
        decorator_list=decorator_list,
        returns=returns,
        type_comment=type_comment,
    )


def update_stubs(path, mod_name, sub_name):
    mod = importlib.import_module(f'{mod_name}')
    namespace = importlib.import_module(f'{mod_name}.{sub_name}')

    ast_root = astor.code_to_ast(namespace)
    # with open(path, "r") as f:
    #     ast_root = ast.parse(f.read())

    # identify classes in the root namespace that are one of the desired types
    target_method_names = '__init__'

    # use the interpreter to identify the names of classes with the desired type
    target_names = [
        name for name, obj in namespace.__dict__.items() if isclass(obj) and issubclass(obj, (ParamAttr, Rack, Device))
    ]

    if len(target_names) > 0:
        print(f'{namespace.__name__}: update {target_names}')
    else:
        return

    # find their definitions in the parsed ast tree
    target_cls_defs = [
        ast_obj for ast_obj in ast.iter_child_nodes(ast_root) if getattr(ast_obj, 'name', None) in target_names
    ]

    for cls_def in target_cls_defs:
        # scrub any existing __init__ stub
        for child in list(cls_def.body):
            if getattr(child, 'name', None) == target_method_names:
                # _prev = child
                cls_def.body.remove(child)
                break

        cls = getattr(namespace, cls_def.name)

        if issubclass(cls, Device):
            attrs = {}

            if 'key' in get_class_attrs(cls):
                attrs['key'] = get_class_attrs(cls)['key']
            attrs.update(
                {
                    name: get_class_attrs(cls)[name]
                    for name in list_value_attrs(cls)
                    if get_class_attrs(cls)[name].sets and name != 'key'
                }
            )

            args = list(attrs.keys())
            defaults = {trait.name: nameit(trait.default) for trait in attrs.values()}
            annotations = {name: nameit(trait._type) for name, trait in attrs.items()}

        elif issubclass(cls, (ParamAttr, Rack)):

            def transform_annot(cls, a):
                if a is T:
                    return cls._type
                else:
                    return a

            raw_annots = getattr(cls, '__annotations__', {})
            raw_annots = {k: v for k, v in raw_annots.items() if not k.startswith('_')}

            if issubclass(cls, Method):
                defaults = dict(key=None)
                annotations = dict(key=None)
                args = list(raw_annots.keys())[::-1]
                args.remove('key')
                args = ['key'] + args
            else:
                defaults = {}
                annotations = {}
                args = list(raw_annots.keys())

            defaults.update({name: getattr(cls, name) for name in args})
            defaults = {name: (None if d is Undefined else d) for name, d in defaults.items()}
            annotations.update({name: transform_annot(cls, type_) for name, type_ in raw_annots.items()})
            annotations = {name: type_ for name, type_ in annotations.items() if type_ not in (Any, None, Undefined)}
        else:
            raise TypeError(f'{cls} is an unknown class type')

        if issubclass(cls, Method) and cls is not Method:
            decorators = [parse_literal('_typing.overload')]

            # for the keyed method determined by setting the 'key' keyword
            annotations['return'] = parse_literal('_bases.TKeyedMethod')
            del defaults['key']
            cls_def.body.insert(
                0,
                ast_function_stub(
                    '__new__', ['cls'] + args, list(defaults.values()), annotations, decorator_list=decorators
                ),
            )

            # for unkeyed (decorator) method
            annotations['return'] = parse_literal('_bases.TDecoratedMethod')
            annotations.pop('key', None)
            args.remove('key')
            cls_def.body.insert(
                0,
                ast_function_stub(
                    '__new__', ['cls'] + args, list(defaults.values()), annotations, decorator_list=decorators
                ),
            )
        else:
            cls_def.body.insert(
                0, ast_function_stub(target_method_names, ['self'] + args, list(defaults.values()), annotations)
            )

    with open(path, 'w') as f:
        f.write(decompile(ast_root))


if __name__ == '__main__':
    root = Path('labbench')
    mod_name = 'labbench'

    # clear out previous files
    for path in root.rglob('*.pyi'):
        Path(path).unlink()

    # stubgen is the first stab
    from mypy import stubgen

    sys.argv = [sys.argv[0], str(root), '-o', str(root / '..'), '-v']
    stubgen.main()

    # now step through to replace __init__ keyword arguments
    for path in root.rglob('*.pyi'):
        if str(path).endswith('notebook.py'):
            continue

        path = Path(path)

        # convert python path to an importable module name
        sub_name = '.'.join(path.with_suffix('').parts[1:])

        update_stubs(path, mod_name=mod_name, sub_name=sub_name)




from . import _bases, _types
import typing as _typing
from ._bases import Method


class any(Method, _types.Any):
    pass


class bool(Method[bool], _types.Bool):
    pass


class float(Method[float], _types.Float):
    pass


class int(Method[int], _types.Int):
    pass


class complex(Method[complex], _types.Complex):
    pass


class str(Method[str], _types.Unicode):
    pass


class bytes(Method, _types.Bytes):
    pass


class list(_bases.Method, _types.List):
    pass


class tuple(_bases.Method, _types.Tuple):
    pass


class dict(_bases.Method, _types.Dict):
    pass


class Path(_bases.Method, _types.Path):
    pass


class NetworkAddress(_bases.Method, _types.NetworkAddress):
    pass




from . import _types
from ._bases import MethodKeywordArgument


class any(MethodKeywordArgument, _types.Any):
    pass


class bool(MethodKeywordArgument, _types.Bool):
    pass


class float(MethodKeywordArgument, _types.Float):
    pass


class int(MethodKeywordArgument, _types.Int):
    pass


class complex(MethodKeywordArgument, _types.Complex):
    pass


class str(MethodKeywordArgument, _types.Unicode):
    pass


class bytes(MethodKeywordArgument, _types.Bytes):
    pass


class list(MethodKeywordArgument, _types.List):
    pass


class tuple(MethodKeywordArgument, _types.Tuple):
    pass


class dict(MethodKeywordArgument, _types.Dict):
    pass


class Path(MethodKeywordArgument, _types.Path):
    pass


class NetworkAddress(MethodKeywordArgument, _types.NetworkAddress):
    pass



import numbers
import typing
from pathlib import Path as _Path

import validators as _val

from .. import util
from . import _bases


class Any(_bases.ParamAttr[typing.Any], type=object):
    """allows any value"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        return value

    @util.hide_in_traceback
    def to_pythonic(self, value):
        return value


class Int(_bases.BoundedNumber[int], type=int):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""


class Float(_bases.BoundedNumber[float], type=float):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""

    step: typing.Union[float, None] = None

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        value = super().validate(value, owner)
        if self.step is not None:
            mod = value % self.step
            if mod < self.step / 2:
                return value - mod
            else:
                return value - (mod - self.step)
        return value

    def doc_params(self, skip: list[str] = ['help', 'label'], as_argument: bool = False) -> str:
        if as_argument:
            return None

        else:
            # for text docs: allow subclasses to document their own params
            docs = []
            if self.step is not None:
                docs.append(f'* Step size: {self.step} {self.label}')

            return '\n'.join(docs) + '\n'


class Complex(_bases.ParamAttr[complex], type=complex):
    """accepts numerical or str values, following normal python casting procedures (with bounds checking)"""

    allow_none: bool = False


class Bool(_bases.ParamAttr[bool], type=bool):
    """accepts boolean or numeric values, or a case-insensitive match to one of ('true',b'true','false',b'false')"""

    allow_none: bool = False

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if isinstance(value, (bool, numbers.Number)):
            return value
        elif isinstance(value, (str, bytes)):
            lvalue = value.lower()
            if lvalue in ('true', b'true'):
                return True
            elif lvalue in ('false', b'false'):
                return False
        raise ValueError(
            f"'{self.__repr__(owner=owner)}' accepts only boolean, numerical values,"
            "or one of ('true',b'true','false',b'false'), case-insensitive"
        )


class String(_bases.ParamAttr):
    """base class for string types, which adds support for case sensitivity arguments"""

    case: bool = True
    # allow_none: bool = True # let's not override this default

    @util.hide_in_traceback
    def contains(self, iterable, value):
        if not self.case:
            iterable = [v.lower() for v in iterable]
            value = value.lower()
        return value in iterable

    def doc_params(self, skip: list[str] = ['help', 'label'], as_argument: bool = False) -> str:
        if as_argument:
            return None

        else:
            # for text docs: allow subclasses to document their own params
            if self.case:
                return '* Case sensitive'
            else:
                return '* Case insensitive'


class Unicode(String[str], type=str):
    """accepts strings or numeric values only; convert others explicitly before assignment"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if not isinstance(value, (str, numbers.Number)):
            raise ValueError(
                f"'{type(self).__qualname__}' attributes accept values of str or numerical type, not {type(value).__name__}"
            )
        return value


class Bytes(String[bytes], type=bytes):
    """accepts bytes objects only - encode str (unicode) explicitly before assignment"""


class Iterable(_bases.ParamAttr):
    """accepts any iterable"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if not hasattr(value, '__iter__'):
            raise ValueError(f"'{type(self).__qualname__}' attributes accept only iterable values")
        return value


class Dict(Iterable[dict], type=dict):
    """accepts any type of iterable value accepted by python `dict()`"""


class List(Iterable[list], type=list):
    """accepts any type of iterable value accepted by python `list()`"""


class Tuple(Iterable[tuple], type=tuple):
    """accepts any type of iterable value accepted by python `tuple()`"""

    sets: bool = False


class Path(_bases.ParamAttr[_Path], type=_Path):
    must_exist: bool = False
    """ does the path need to exist when set? """

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        path = self._type(value)

        if self.must_exist and not path.exists():
            raise OSError('no file at path {value}')

        return path

    def doc_params(self, skip: list[str] = ['help', 'label'], as_argument: bool = False) -> str:
        if as_argument:
            return None

        else:
            # for text docs: allow subclasses to document their own params
            if self.must_exist:
                return '* Path name must exist on the host'


class NetworkAddress(Unicode):
    """a IDN-compatible network address string, such as an IP address or DNS hostname"""

    accept_port: bool = True

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        """Rough IDN compatible domain validator"""

        host, *extra = value.split(':', 1)

        if len(extra) > 0:
            port = extra[0]
            try:
                int(port)
            except ValueError:
                raise ValueError(f'port {port} in "{value}" is invalid')

            if not self.accept_port:
                raise ValueError(f'{self} does not accept a port number (accept_port=False)')

        for validate in _val.ipv4, _val.ipv6, _val.domain, _val.slug:
            if validate(host):
                break
        else:
            raise ValueError('invalid host address')

        return value



from numbers import Number

import pytest
from paramattr_tooling import eval_set_then_get, has_steps, loop_closed_loop_set_gets

import labbench as lb
from labbench import paramattr as attr
from labbench.testing import store_backend


@store_backend.key_adapter(defaults={'str_or_none': None, 'str_cached': 'cached string'})
class StoreTestDevice(store_backend.StoreTestDevice):
    LOOP_TEST_VALUES = {
        # make sure all test values conform to these general test values
        int: 5,
        float: 3.14,
        str: 'moose',
        bool: True,
        object: None,
    }

    # test both getting and setting
    bool: bool = attr.value.bool(default=True)
    int_with_default = attr.value.int(default=47)
    int_no_default = attr.value.int(default=None, allow_none=True)

    # floats
    float_low_bounded = attr.value.float(default=2, min=1)
    float_high_bounded = attr.value.float(default=2, max=4)
    float_none_bounded = attr.value.float(default=3, allow_none=True, max=4, min=1)
    float_stepped = attr.value.float(default=0, step=3)

    # strings
    str_explicit_none = attr.value.str(default=None, allow_none=True)
    str_allow_none = attr.value.str(default=None, allow_none=True)
    str_cached = attr.value.str(default='47', cache=True)
    str = attr.value.str(default='squirrel')
    any = attr.value.any(default='empty', allow_none=True)
    str_with_only = attr.value.str(default='moose', only=('moose', 'squirrel'))
    str_no_case_with_only = attr.value.str(default='moose', only=('MOOSE', 'squirrel'), case=False)


@attr.adjust('bool', default=False)
class AdjustedTestDevice(StoreTestDevice):
    pass


class RequiredParametersTestDevice(lb.Device):
    required_str: str = attr.value.str(cache=True)
    required_str_allow_none: str = attr.value.str(cache=True, allow_none=True)


def change_case(s: str):
    alt = s.upper()
    if alt == s:
        alt = s.lower()
    return alt


def set_param(device, attr_name, value, arguments={}):
    if len(arguments) > 0:
        attr_def = attr.get_class_attrs(device)[attr_name]
        raise ValueError(f'{attr_def.ROLE} do not accept arguments')
    setattr(device, attr_name, value)


def get_param(device, attr_name, arguments={}):
    if len(arguments) > 0:
        attr_def = attr.get_class_attrs(device)[attr_name]
        raise ValueError(f'{attr_def.ROLE} properties do not accept arguments')
    return getattr(device, attr_name)


def set_then_get(device, attr_name, value_in, arguments={}):
    set_param(device, attr_name, value_in, arguments)
    return get_param(device, attr_name, arguments)


#
# Fixtures that convert to arguments in test functions
#
@pytest.fixture(autouse=True, scope='module')
def labbench_fixture():
    lb.visa_default_resource_manager('@sim')
    lb.show_messages('info')
    lb.util.force_full_traceback(True)


@pytest.fixture
def role_type():
    return lb.paramattr.value.Value


@pytest.fixture
def opened_device():
    device = StoreTestDevice()
    device.open()
    yield device
    device.close()


@pytest.fixture
def opened_adjusted_device():
    device = AdjustedTestDevice()
    device.open()
    yield device
    device.close()


@pytest.fixture
def instantiated_device():
    device = StoreTestDevice()
    device.open()
    yield device
    device.close()


#
# The tests
#
def test_basic_get(opened_device):
    get_param(opened_device, 'any')


def test_basic_set(opened_device):
    opened_device.any = 5
    assert opened_device.any == 5


def test_cache(opened_device):
    # repeat to set->get to ensure proper caching
    eval_set_then_get(opened_device, 'str_cached', set_then_get)
    result = eval_set_then_get(opened_device, 'str_cached', set_then_get)
    assert len(result['notifications']) == 2, 'notification count for cached string'


def test_default_types(opened_device, role_type):
    for attr_def in opened_device.get_attr_defs().values():
        if not isinstance(attr_def, role_type):
            continue

        value = getattr(opened_device, attr_def.name)

        if attr_def.allow_none:
            allow_types = (type(None), attr_def._type)
        else:
            allow_types = (attr_def._type,)

        if issubclass(attr_def._type, Number):
            allow_types = allow_types + (Number,)

        assert issubclass(type(value), allow_types), f'pythonic type of {attr_def.name}'


def test_default_values(opened_device, role_type):
    for attr_def in opened_device.get_attr_defs().values():
        if not isinstance(attr_def, role_type):
            continue

        value = getattr(opened_device, attr_def.name)
        assert value == attr_def.default, f'pythonic type of {attr_def.name}'


def test_only(opened_device):
    # low bound
    expected_valid = type(opened_device).str_with_only.only[0]
    alt_case = change_case(expected_valid)

    opened_device.str_with_only = expected_valid
    with pytest.raises(ValueError):
        opened_device.str_with_only = 'boris'
    with pytest.raises(ValueError):
        opened_device.str_with_only = alt_case

    # with string case
    expected_valid = type(opened_device).str_no_case_with_only.only[0]
    alt_case = change_case(expected_valid)

    opened_device.str_no_case_with_only = expected_valid
    with pytest.raises(ValueError):
        opened_device.str_no_case_with_only = 'boris'
    opened_device.str_no_case_with_only = alt_case


def test_numeric_bounds(opened_device):
    # float_low_bounded
    lo_bound = type(opened_device).float_low_bounded.min
    with pytest.raises(ValueError):
        opened_device.float_low_bounded = lo_bound - 1
    opened_device.float_low_bounded = lo_bound + 1

    # float_high_bounded
    hi_bound = type(opened_device).float_high_bounded.max
    with pytest.raises(ValueError):
        opened_device.float_high_bounded = hi_bound + 1
    opened_device.float_high_bounded = hi_bound - 1

    # float_none_bounded
    opened_device.float_none_bounded = None


def test_numeric_casting(opened_device):
    # float
    value_in = '3.91'
    expected_out = float(value_in)
    opened_device.float_low_bounded = value_in
    value_out = opened_device.float_low_bounded
    assert value_out == expected_out, 'string to float casting'

    # float
    value_in = '-48'
    expected_out = int(value_in)
    opened_device.int_no_default = value_in
    value_out = opened_device.int_no_default
    assert value_out == expected_out, 'string to float casting'


def test_str_casting(opened_device):
    # float
    value_in = -48
    expected_out = str(value_in)
    opened_device.str = value_in
    value_out = opened_device.str
    assert value_out == expected_out, 'string to float casting'


def test_numeric_step(opened_device):
    # rounding tests (step is not None)
    opened_device.float_stepped = 3
    assert opened_device.float_stepped == 3.0
    opened_device.float_stepped = 2
    assert opened_device.float_stepped == 3.0
    opened_device.float_stepped = 4
    assert opened_device.float_stepped == 3.0
    opened_device.float_stepped = 1.6
    assert opened_device.float_stepped == 3.0
    opened_device.float_stepped = -2
    assert opened_device.float_stepped == -3.0
    opened_device.float_stepped = -1
    assert opened_device.float_stepped == 0


def test_device_initialization(instantiated_device, role_type):
    def should_test_this_attr(attr_def: attr.ParamAttr):
        return (
            attr_def.name in type(instantiated_device).__annotations__
            and isinstance(attr_def, role_type)
            and attr_def.sets
            and not has_steps(attr_def)
        )

    cls = type(instantiated_device)

    for attr_def in instantiated_device.get_attr_defs().values():
        if not should_test_this_attr(attr_def):
            continue

        test_name = f'{attr_def.ROLE} "{attr_def.name}"'

        value_in = instantiated_device.LOOP_TEST_VALUES[attr_def._type]

        instantiated_device = cls(**{attr_def.name: value_in})

        with instantiated_device:
            value_out = getattr(instantiated_device, attr_def.name)
            assert value_in == value_out, f'{test_name} - initialize default values from Device constructor'


def test_all_get_sets(opened_device, role_type):
    loop_closed_loop_set_gets(opened_device, role_type, set_then_get)


def test_adjusted_value_paramattr(opened_adjusted_device):
    assert opened_adjusted_device.bool == False


def test_requried_parameter_instantiation():
    with pytest.raises(TypeError):
        RequiredParametersTestDevice()

    RequiredParametersTestDevice(required_str='hi', required_str_allow_none=None)



# -*- coding: UTF-8 -*-

# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

import time

import numpy as np
import pytest

import labbench as lb
from labbench import paramattr as attr
from labbench.testing.store_backend import StoreTestDevice


class LaggyInstrument(StoreTestDevice):
    """A mock "instrument" to measure time response in (a)sync operations"""

    delay: float = attr.value.float(default=0, min=0, help='connection time')
    fetch_time: float = attr.value.float(default=0, min=0, help='fetch time')
    fail_disconnect = attr.value.bool(default=False, help='whether to raise DivideByZero on disconnect')

    def open(self):
        self.perf = {}
        self._logger.info(f'{self} connect start')
        t0 = time.perf_counter()
        lb.sleep(self.delay)
        self.perf['open'] = time.perf_counter() - t0
        self._logger.info(f'{self} connected')

    def fetch(self):
        """Return the argument after a 1s delay"""
        lb.logger.info(f'{self}.fetch start')
        t0 = time.perf_counter()
        lb.sleep(self.fetch_time)
        lb.logger.info(f'{self}.fetch done')
        self.perf['fetch'] = time.perf_counter() - t0
        return self.fetch_time

    def dict(self):
        return {self.resource: self.resource}

    def none(self):
        """Return None"""
        return None

    def close(self):
        self._logger.info(f'{self} disconnected')
        if self.fail_disconnect:
            1 / 0


class Rack1(lb.Rack):
    dev1: LaggyInstrument = LaggyInstrument()
    dev2: LaggyInstrument = LaggyInstrument()

    def setup(self, param1: float):
        pass

    def arm(self):
        self.dev1.variable_setting = self.dev1.variable_setting + 2

    def open(self):
        time.sleep(0.25)


class Rack2(lb.Rack):
    dev: LaggyInstrument = LaggyInstrument()

    def setup(self):
        return 'rack 2 - setup'
        return self.dev.dict()

    def acquire(self, *, param1):
        return 'rack 3 - acquire'

    def fetch(self, *, param2: int = 7):
        return self.dev.fetch()

    def open(self):
        time.sleep(0.25)


class Rack3(lb.Rack):
    # this is unset (no =), so it _must_ be passed as an argument to instantiate, i.e.,
    #
    # >>> Rack3(dev=MyDevice)
    dev: LaggyInstrument

    # notice this operation requires param2 and param3
    def acquire(self, *, param3, param2=7):
        pass

    def fetch(self, *, param4):
        return self.dev.fetch()

    def open(self):
        time.sleep(0.25)


class MyRack(lb.Rack):
    db = lb.CSVLogger(
        path=time.strftime(
            'test db/test-rack %Y-%m-%d_%Hh%Mm%Ss'
        ),  # Path to new directory that will contain containing all files
        append=True,  # `True` --- allow appends to an existing database; `False` --- append
        text_relational_min=1024,  # Minimum text string length that triggers relational storage
        force_relational=['host_log'],  # Data in these columns will always be relational
        nonscalar_file_type='csv',  # Default format of numerical data, when possible
        tar=False,  # `True` to embed relational data folders within `data.tar`
    )

    # Devices
    inst1: LaggyInstrument = LaggyInstrument(resource='a', delay=0.12)
    inst2: LaggyInstrument = LaggyInstrument(resource='b', delay=0.06)

    # Test procedures
    rack1 = Rack1(dev1=inst1, dev2=inst2)
    rack2 = Rack2(dev=inst1)
    rack3 = Rack3(dev=inst2)

    run = lb.Sequence(
        (rack1.setup, rack2.setup),  # 2 methods run concurrently
        rack1.arm,
        rack2.acquire,
        rack3.acquire,
        (rack2.fetch, rack3.fetch),  # 2 also run concurrently
        db.new_row,
    )


@pytest.fixture
def db_path():
    return f'test db/{np.random.bytes(8).hex()}'


def test_context_open():
    rack = MyRack()

    assert not rack.inst1.isopen
    assert not rack.inst2.isopen

    with MyRack():
        assert rack.inst1.isopen
        assert rack.inst2.isopen

    assert not rack.inst1.isopen
    assert not rack.inst2.isopen


def test_call_open():
    rack = MyRack()

    assert not rack.inst1.isopen
    assert not rack.inst2.isopen

    rack.open()
    try:
        assert rack.inst1.isopen
        assert rack.inst2.isopen
    finally:
        rack.close()

    assert not rack.inst1.isopen
    assert not rack.inst2.isopen


# if __name__ == "__main__":
# rack = MyRack()
# rack.open()

# # Testbed = lb.Rack.take_module('module_as_testbed')
# Testbed = MyRack

# lst = CommentedSeq(['a', 'b'])
# lst.yaml_add_eol_comment("foo", 0, 0)
# lst.yaml_add_eol_comment("bar\n\n", 1)
# data["list_of_elements_side_comment"] = lst
# data.yaml_set_comment_before_after_key("list_of_elements_side_comment", "\n")

# lst = CS(['a', 'b'])
# lst.yaml_set_comment_before_after_key(0, "comment 1", 2)
# lst.yaml_set_comment_before_after_key(1, "comment 2", 2)
# data["list_of_elements_top_comment"] = lst

# Testbed.config.make_templates()

# # Testbed.config._rack_defaults(Testbed)

# with lb.stopwatch('test connection'):
#     with Testbed() as testbed:
#         # with MyRack() as testbed:
#         testbed.inst2.delay = 0.07
#         testbed.inst1.delay = 0.12

#         testbed.run.iterate_from_csv('setup/MyRack.run.csv')

#         # for i in range(3):
#         #     # Run the experiment
#         #     ret = testbed.run(rack1_param1=1, rack2_param1=2, rack3_param2=3,
#         #                       rack3_param3=4, rack2_param2=5, rack3_param4=6)
#         #
#         #     testbed.db()

# df = lb.read(testbed.db.path/'master.db')
# df.to_csv(testbed.db.path/'master.csv')

# config_root = Path('test-config')
# config_root.mkdir(parents=True, exist_ok=True)

# # default devices
# defaults = {
#     dev_name: {k:getattr(dev, k) for k in dev._value_attrs}
#     for dev_name, dev in MyRack._devices.items()
# }
# with open(config_root/'devices.yaml', 'w') as f:
#     yaml.dump(defaults, f)

# defaults, annots, methods, names = MyRack.config.parameters(MyRack)

# with open(config_root/'rack arguments.yaml', 'w') as f:
#     f.write('## Comment or remove lines to add as columns in sequence tables\n\n')
#     for k, default in defaults.items():
#         if default is lb._rack.EMPTY:
#             s = f'# {k}: \n'
#             f.write(s)
#         else:
#             s = yaml.dump({k: default}, f)
#         if annots[k] is not lb._rack.EMPTY:
#             before, *after = s.split(b'\n', 1)
#             s = b'\n'.join([before+f" # {annots[k].__qualname__}"] + after)



import importlib

from test_sequencing import LaggyInstrument

import labbench as lb

lb = importlib.reload(lb)


class Rack1(lb.Rack):
    dev1: lb.Device = LaggyInstrument()
    dev2: lb.Device = LaggyInstrument()

    def setup(self, param1):
        pass

    def arm(self):
        pass


class Rack2(lb.Rack):
    dev: lb.Device = LaggyInstrument()

    def setup(self):
        return 'rack 2 - setup'
        return self.dev.dict()

    def acquire(self, *, param1):
        return 'rack 3 - acquire'

    def fetch(self, *, param2=7):
        return self.dev.fetch()


class Rack3(lb.Rack):
    dev: lb.Device = LaggyInstrument()

    def acquire(self, *, param2=7, param3):
        pass

    def fetch(self, *, param4):
        self.dev.fetch()


db: lb._data.TabularLoggerBase = lb.SQLiteLogger(
    'data',  # Path to new directory that will contain containing all files
    append=True,  # `True` --- allow appends to an existing database; `False` --- append
    text_relational_min=1024,  # Minimum text string length that triggers relational storage
    force_relational=['host_log'],  # Data in these columns will always be relational
    nonscalar_file_type='csv',  # Default format of numerical data, when possible
    metadata_dirname='metadata',  # metadata will be stored in this subdirectory
    tar=False,  # `True` to embed relational data folders within `data.tar`
)

# Devices
inst1: LaggyInstrument = LaggyInstrument(resource='a', delay=0.12)
inst2: LaggyInstrument = LaggyInstrument(resource='b', delay=0.06)

# Test procedures
rack1 = Rack1(dev1=inst1, dev2=inst2)
rack2 = Rack2(dev=inst1)
rack3 = Rack3(dev=inst2)

run = lb.Sequence(
    (rack1.setup, rack2.setup),  # executes these 2 methods concurrently
    (rack1.arm),
    rack2.acquire,
    rack3.acquire,  # executes these 2 sequentially
    (rack2.fetch, rack3.fetch),
    (db.new_row),
)



import pytest

import labbench as lb
from labbench.testing import pyvisa_sim


@pytest.fixture(autouse=True, scope='module')
def setup_labbench():
    lb.visa_default_resource_manager('@sim')
    lb.util.force_full_traceback(True)


def test_adjust():
    scope = pyvisa_sim.Oscilloscope()
    specan = pyvisa_sim.SpectrumAnalyzer()

    assert scope.model != specan.model


def test_visadevice_no_resource():
    inst = lb.VISADevice()

    with pytest.raises(ConnectionError):
        inst.open()


def test_visadevice_explicit_resource():
    inst = lb.VISADevice('USB::0x1111::0x2233::0x9876::INSTR')
    with inst:
        pass


def test_visadevice_serial_resource():
    inst_explicit = lb.VISADevice('USB::0x1111::0x2233::0x9876::INSTR')
    inst_auto = lb.VISADevice('OMGBBQ58')
    with inst_auto, inst_explicit:
        assert inst_auto._identity == inst_explicit._identity


def test_subclass_explicit_resource():
    inst = pyvisa_sim.Oscilloscope('USB::0x1111::0x2233::0x9876::INSTR')
    with inst:
        pass


def test_subclass_no_resource():
    inst_explicit = lb.VISADevice('USB::0x1111::0x2233::0x9876::INSTR')
    inst_auto = pyvisa_sim.Oscilloscope()
    with inst_auto, inst_explicit:
        assert inst_auto._identity == inst_explicit._identity


def test_subclass_serial_resource():
    inst_explicit = lb.VISADevice('USB::0x1111::0x2233::0x9876::INSTR')
    inst_auto = pyvisa_sim.Oscilloscope('OMGBBQ58')
    with inst_auto, inst_explicit:
        assert inst_auto._identity == inst_explicit._identity


def test_method_access():
    inst = pyvisa_sim.Oscilloscope()

    with inst:
        inst.resolution_bandwidth(10e3, channel=2)
        inst.resolution_bandwidth(channel=1)
